#!/usr/bin/env python

# ------------------------------------------------------------------------------

# VIRTUAL ROBOT

# Copyright (C) 2016
# Mike Maready
# All Rights Reserved

# This program was designed and developed by Mike Maready. Permission
# is granted to use this software for non-commercial purposes only.
# You may use, modify, and distribute this software freely without any
# other restrictions as long as this section of text stays in the
# source and all associated files are kept together. If you modify and
# distribute this software, you must do so under these same terms.

# No warranty expressed or implied (you get what you pay for).

# ------------------------------------------------------------------------------

# Revision History:

# 02/12/2016 : Demo Version
# 03/11/2016 : Version 1.0

# See changelog.txt for details.

# ------------------------------------------------------------------------------

# imports

import sys
import pygame

from math import sin
from math import cos
from math import atan
from math import atan2
from math import sqrt

# ------------------------------------------------------------------------------

# globals

collision_detection = 0
obstacle_avoidance_mode = 0

pi = 3.141592653589

# ------------------------------------------------------------------------------

# robot class

class Robot(pygame.sprite.Sprite):

   def __init__(self, xpos, ypos, dir):
      pygame.sprite.Sprite.__init__(self)
      self.imageMaster = pygame.image.load("robot.bmp")
      self.imageMaster.set_colorkey((255, 255, 255))
      self.imageMaster = self.imageMaster.convert_alpha()
      self.image = self.imageMaster
      self.rect = self.image.get_rect()
      self.rect.center = (xpos, ypos)
      self.mask = pygame.mask.from_surface(self.image)
      self.resetx = xpos
      self.resety = ypos
      self.resetd = dir
      self.x = xpos
      self.y = ypos
      self.dir = dir

   def update(self):
      oldCenter = self.rect.center
      self.image = pygame.transform.rotate(self.imageMaster, self.dir)
      self.rect = self.image.get_rect()
      self.rect.center = oldCenter
      self.mask = pygame.mask.from_surface(self.image)

# ------------------------------------------------------------------------------

# obstacle class

class Obstacle(pygame.sprite.Sprite):

   def __init__(self, xpos, ypos, dir, file):
      pygame.sprite.Sprite.__init__(self)
      self.imageMaster = pygame.image.load(file)
      self.imageMaster.set_colorkey((255, 255, 255))
      self.imageMaster = self.imageMaster.convert_alpha()
      self.image = self.imageMaster
      self.rect = self.image.get_rect()
      self.rect.center = (xpos, ypos)
      self.mask = pygame.mask.from_surface(self.image)
      self.dir = dir

   def update(self):
      oldCenter = self.rect.center
      self.image = pygame.transform.rotate(self.imageMaster, self.dir)
      self.rect = self.image.get_rect()
      self.rect.center = oldCenter
      self.mask = pygame.mask.from_surface(self.image)

# ------------------------------------------------------------------------------

# goal class

class Goal(pygame.sprite.Sprite):

   def __init__(self, xpos, ypos):
      pygame.sprite.Sprite.__init__(self)
      self.image = pygame.Surface([11, 11])
      self.image.fill((0, 200, 0))
      self.rect = self.image.get_rect()
      self.resetx = xpos - 5
      self.resety = ypos - 5
      self.rect.x = xpos - 5
      self.rect.y = ypos - 5

# ------------------------------------------------------------------------------

# update screen

def update_screen():

   all_sprites_list.clear(screen, background)
   all_sprites_list.update()
   all_sprites_list.draw(screen)
   pygame.display.flip()

   return 0

# ------------------------------------------------------------------------------

# help

def help():

   print "Available Commands:"
   print "? = Show This Help"
   print "keyboard = Keyboard Mode"
   print "mouse 1 = Mouse Mode 1"
   print "mouse 2 = Mouse Mode 2"
   print "rs FILE = Run Script"
   print "robot X Y DIR = Set Robot Defaults"
   print "obstacle1 X Y DIR FILE = Set Obstacle 1 Defaults"
   print "obstacle2 X Y DIR FILE = Set Obstacle 2 Defaults"
   print "goal X Y = Set Goal Defaults"
   print "rpd ? = Get Robot Position And Direction"
   print "rpd X Y DIR = Set Robot Position And Direction"
   print "rp ? = Get Robot Position"
   print "rp X Y = Set Robot Position"
   print "rd ? = Get Robot Direction"
   print "rd DIR = Set Robot Direction"
   print "gp ? = Get Goal Position"
   print "gp X Y = Set Goal Position"
   print "cd ? = Get Collision Detection"
   print "cd on = Set Collision Detection = On"
   print "cd off = Set Collision Detection = Off"
   print "oam ? = Get Obstacle Avoidance Mode"
   print "oam 0 = Set Obstacle Avoidance Mode = 0"
   print "oam 1 = Set Obstacle Avoidance Mode = 1"
   print "oam 2 = Set Obstacle Avoidance Mode = 2"
   print "cdd X Y = Calculate Direction And Distance"
   print "rl DEG = Rotate Left"
   print "rr DEG = Rotate Right"
   print "rtd DIR = Rotate To Direction"
   print "rtp X Y = Rotate To Position"
   print "mf DIST = Move Forward"
   print "mb DIST = Move Backward"
   print "mtp X Y = Move To Position"
   print "mtg = Move To Goal"
   print "pause = Pause Program"
   print "r = Reset"
   print "q = Quit"

   return 0

# ------------------------------------------------------------------------------

# keyboard mode

def keyboard_mode():

   print "Keyboard Mode"
   print "D = Rotate Left"
   print "F = Rotate Right"
   print "J = Move Forward"
   print "K = Move Backward"
   print "Esc = Done"

   keep_going = True

   while keep_going:

      for event in pygame.event.get():
         if event.type == pygame.QUIT:
            keep_going = False
         elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
               keep_going = False
            elif event.key == pygame.K_d:
               robot.dir += 2
            elif event.key == pygame.K_f:
               robot.dir -= 2
            elif event.key == pygame.K_j:
               d = robot.dir
               rad = d * pi / 180
               robot.x -= 2 * sin(rad)
               robot.y -= 2 * cos(rad)
               robot.rect.center = (robot.x, robot.y)
            elif event.key == pygame.K_k:
               d = robot.dir
               rad = d * pi / 180
               robot.x += 2 * sin(rad)
               robot.y += 2 * cos(rad)
               robot.rect.center = (robot.x, robot.y)

      robot.dir = normalize(robot.dir)

      update_screen()

      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         if collision_detection == 1:
            print "COLLISION"
            keep_going = False

      clock.tick(30) # ?

   return 0

# ------------------------------------------------------------------------------

# mouse mode 1

def mouse_mode_1():

   print "Mouse Mode 1"
   print "Move Mouse = Move Robot"
   print "Scroll Up = Rotate Left"
   print "Scroll Down = Rotate Right"
   print "D = Rotate Left"
   print "F = Rotate Right"
   print "Esc = Done"

   keep_going = True

   while keep_going:

      for event in pygame.event.get():
         if event.type == pygame.QUIT:
            keep_going = False
         elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
               keep_going = False
            elif event.key == pygame.K_d:
               robot.dir += 2
            elif event.key == pygame.K_f:
               robot.dir -= 2
         if event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == 4: # scroll up
               robot.dir += 2
            elif event.button == 5: # scroll down
               robot.dir -= 2

      pos = pygame.mouse.get_pos()
      robot.x = pos[0]
      robot.y = pos[1]
      robot.rect.center = (robot.x, robot.y)

      robot.dir = normalize(robot.dir)

      update_screen()

      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         if collision_detection == 1:
            print "COLLISION"
            keep_going = False

      clock.tick(30) # ?

   return 0

# ------------------------------------------------------------------------------

# mouse mode 2

def mouse_mode_2():

   print "Mouse Mode 2"
   print "Left Click = Set Robot Position"
   print "Right Click = Set Goal Position"
   print "Scroll Up = Rotate Left"
   print "Scroll Down = Rotate Right"
   print "D = Rotate Left"
   print "F = Rotate Right"
   print "Esc = Done"

   keep_going = True

   while keep_going:

      for event in pygame.event.get():
         if event.type == pygame.QUIT:
            keep_going = False
         elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
               keep_going = False
            elif event.key == pygame.K_d:
               robot.dir += 2
            elif event.key == pygame.K_f:
               robot.dir -= 2
         elif event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == 1: # left button
               pos = pygame.mouse.get_pos()
               robot.x = pos[0]
               robot.y = pos[1]
               robot.rect.center = (robot.x, robot.y)
            elif event.button == 3: # right button
               pos = pygame.mouse.get_pos()
               goal.rect.center = (pos[0], pos[1])
            elif event.button == 4: # scroll up
               robot.dir += 2
            elif event.button == 5: # scroll down
               robot.dir -= 2

      robot.dir = normalize(robot.dir)

      update_screen()

      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         if collision_detection == 1:
            print "COLLISION"
            keep_going = False

      clock.tick(30) # ?

   return 0

# ------------------------------------------------------------------------------

# run script

def run_script(file):

   print "Run Script"
   try:
      f = open(file, 'r')
   except IOError:
      print "Error (IOError)."
      return 1
   for line in f:
      do_input(line.strip())
   f.close()

   return 0

# ------------------------------------------------------------------------------

# do input

def do_input(input):

   words = input.split()

   # robot X Y DIR

   for not_used in range(0, 1):
      if len(words) != 4:
         break
      if words[0] != 'robot':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      if not is_float(words[3]):
         break
      x = int(words[1])
      y = int(words[2])
      d = float(words[3])
      if d < -360 or d > 360:
         break
      set_robot_defaults(x, y, d)

   # obstacle1 X Y DIR FILE

   for not_used in range(0, 1):
      if len(words) != 5:
         break
      if words[0] != 'obstacle1':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      if not is_float(words[3]):
         break
      x = int(words[1])
      y = int(words[2])
      d = float(words[3])
      if d < -360 or d > 360:
         break
      set_obstacle_1_defaults(x, y, d, words[4])

   # obstacle2 X Y DIR FILE

   for not_used in range(0, 1):
      if len(words) != 5:
         break
      if words[0] != 'obstacle2':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      if not is_float(words[3]):
         break
      x = int(words[1])
      y = int(words[2])
      d = float(words[3])
      if d < -360 or d > 360:
         break
      set_obstacle_2_defaults(x, y, d, words[4])

   # goal X Y

   for not_used in range(0, 1):
      if len(words) != 3:
         break
      if words[0] != 'goal':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      x = int(words[1])
      y = int(words[2])
      set_goal_defaults(x, y)

   # rpd ?

   if input == 'rpd ?':
      get_robot_position_and_direction()

   # rpd X Y DIR

   for not_used in range(0, 1):
      if len(words) != 4:
         break
      if words[0] != 'rpd':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      if not is_float(words[3]):
         break
      x = int(words[1])
      y = int(words[2])
      d = float(words[3])
      if d < -360 or d > 360:
         break
      set_robot_position_and_direction(x, y, d)

   # rp ?

   if input == 'rp ?':
      get_robot_position()

   # rp X Y

   for not_used in range(0, 1):
      if len(words) != 3:
         break
      if words[0] != 'rp':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      x = int(words[1])
      y = int(words[2])
      set_robot_position(x, y)

   # rd ?

   if input == 'rd ?':
      get_robot_direction()

   # rd DIR

   for not_used in range(0, 1):
      if len(words) != 2:
         break
      if words[0] != 'rd':
         break
      if not is_float(words[1]):
         break
      d = float(words[1])
      if d < -360 or d > 360:
         break
      set_robot_direction(d)

   # gp ?

   if input == 'gp ?':
      get_goal_position()

   # gp X Y

   for not_used in range(0, 1):
      if len(words) != 3:
         break
      if words[0] != 'gp':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      x = int(words[1])
      y = int(words[2])
      set_goal_position(x, y)

   # cd ?

   if input == 'cd ?':
      get_collision_detection()

   # cd on

   if input == 'cd on':
      set_collision_detection(1)

   # cd off

   if input == 'cd off':
      set_collision_detection(0)

   # oam ?

   if input == 'oam ?':
      get_obstacle_avoidance_mode()

   # oam 0

   if input == 'oam 0':
      set_obstacle_avoidance_mode(0)

   # oam 1

   if input == 'oam 1':
      set_obstacle_avoidance_mode(1)

   # oam 2

   if input == 'oam 2':
      set_obstacle_avoidance_mode(2)

   # cdd X Y

   for not_used in range(0, 1):
      if len(words) != 3:
         break
      if words[0] != 'cdd':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      x = int(words[1])
      y = int(words[2])
      calculate_direction_and_distance(x, y)

   # rl DEG

   for not_used in range(0, 1):
      if len(words) != 2:
         break
      if words[0] != 'rl':
         break
      if not is_float(words[1]):
         break
      d = float(words[1])
      if d < -360 or d > 360:
         break
      rotate_left(d)

   # rr DEG

   for not_used in range(0, 1):
      if len(words) != 2:
         break
      if words[0] != 'rr':
         break
      if not is_float(words[1]):
         break
      d = float(words[1])
      if d < -360 or d > 360:
         break
      rotate_right(d)

   # rtd DIR

   for not_used in range(0, 1):
      if len(words) != 2:
         break
      if words[0] != 'rtd':
         break
      if not is_float(words[1]):
         break
      d = float(words[1])
      if d < -360 or d > 360:
         break
      rotate_to_direction(d)

   # rtp X Y

   for not_used in range(0, 1):
      if len(words) != 3:
         break
      if words[0] != 'rtp':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      x = int(words[1])
      y = int(words[2])
      rotate_to_position(x, y)

   # mf DIST

   for not_used in range(0, 1):
      if len(words) != 2:
         break
      if words[0] != 'mf':
         break
      if not is_float(words[1]):
         break
      d = float(words[1])
      move_forward(d)

   # mb DIST

   for not_used in range(0, 1):
      if len(words) != 2:
         break
      if words[0] != 'mb':
         break
      if not is_float(words[1]):
         break
      d = float(words[1])
      move_backward(d)

   # mtp X Y

   for not_used in range(0, 1):
      if len(words) != 3:
         break
      if words[0] != 'mtp':
         break
      if not words[1].isdigit():
         break
      if not words[2].isdigit():
         break
      x = int(words[1])
      y = int(words[2])
      move_to_position(x, y)

   # mtg

   if input == 'mtg':
      move_to_goal()

   # pause

   if input == 'pause':
      pause_program()

   # r

   if input == 'r':
      reset()

   return 0

# ------------------------------------------------------------------------------

# set robot defaults

def set_robot_defaults(xpos, ypos, dir):

   print "Set Robot Defaults"
   dir = normalize(dir)
   robot.resetx = xpos
   robot.resety = ypos
   robot.resetd = dir
   robot.x = robot.resetx
   robot.y = robot.resety
   robot.dir = robot.resetd
   robot.rect.center = (robot.x, robot.y)
   update_screen()
   print "Position = %.2f %.2f" % (robot.x, robot.y)
   print "Direction = %.2f" % robot.dir

   return 0

# ------------------------------------------------------------------------------

# set obstacle 1 defaults

def set_obstacle_1_defaults(xpos, ypos, dir, file):

   global obstacle1

   print "Set Obstacle 1 Defaults"
   dir = normalize(dir)
   all_sprites_list.remove(obstacle1)
   obstacles_list.remove(obstacle1)
   obstacle1 = None
   obstacle1 = Obstacle(xpos, ypos, dir, file)
   all_sprites_list.add(obstacle1)
   obstacles_list.add(obstacle1)
   update_screen()

   return 0

# ------------------------------------------------------------------------------

# set obstacle 2 defaults

def set_obstacle_2_defaults(xpos, ypos, dir, file):

   global obstacle2

   print "Set Obstacle 2 Defaults"
   dir = normalize(dir)
   all_sprites_list.remove(obstacle2)
   obstacles_list.remove(obstacle2)
   obstacle2 = None
   obstacle2 = Obstacle(xpos, ypos, dir, file)
   all_sprites_list.add(obstacle2)
   obstacles_list.add(obstacle2)
   update_screen()

   return 0

# ------------------------------------------------------------------------------

# set goal defaults

def set_goal_defaults(xpos, ypos):

   print "Set Goal Defaults"
   goal.resetx = xpos - 5
   goal.resety = ypos - 5
   goal.rect.x = goal.resetx
   goal.rect.y = goal.resety
   update_screen()
   print "Position = %.2f %.2f" % (goal.rect.x, goal.rect.y)

   return 0

# ------------------------------------------------------------------------------

# get robot position and direction

def get_robot_position_and_direction():

   print "Get Robot Position And Direction"
   print "Position = %.2f %.2f" % (robot.x, robot.y)
   print "Direction = %.2f" % robot.dir

   return 0

# ------------------------------------------------------------------------------

# set robot position and direction

def set_robot_position_and_direction(xpos, ypos, dir):

   print "Set Robot Position And Direction"
   dir = normalize(dir)
   robot.x = xpos
   robot.y = ypos
   robot.dir = dir
   robot.rect.center = (robot.x, robot.y)
   update_screen()
   print "Position = %.2f %.2f" % (robot.x, robot.y)
   print "Direction = %.2f" % robot.dir

   return 0

# ------------------------------------------------------------------------------

# get robot position

def get_robot_position():

   print "Get Robot Position"
   print "Position = %.2f %.2f" % (robot.x, robot.y)

   return 0

# ------------------------------------------------------------------------------

# set robot position

def set_robot_position(xpos, ypos):

   print "Set Robot Position"
   robot.x = xpos
   robot.y = ypos
   robot.rect.center = (robot.x, robot.y)
   update_screen()
   print "Position = %.2f %.2f" % (robot.x, robot.y)

   return 0

# ------------------------------------------------------------------------------

# get robot direction

def get_robot_direction():

   print "Get Robot Direction"
   print "Direction = %.2f" % robot.dir

   return 0

# ------------------------------------------------------------------------------

# set robot direction

def set_robot_direction(dir):

   print "Set Robot Direction"
   dir = normalize(dir)
   robot.dir = dir
   update_screen()
   print "Direction = %.2f" % robot.dir

   return 0

# ------------------------------------------------------------------------------

# get goal position

def get_goal_position():

   print "Get Goal Position"
   xpos = goal.rect.x + 5
   ypos = goal.rect.y + 5
   print "Position = %.2f %.2f" % (xpos, ypos)

   return 0

# ------------------------------------------------------------------------------

# set goal position

def set_goal_position(xpos, ypos):

   print "Set Goal Position"
   goal.rect.x = xpos - 5
   goal.rect.y = ypos - 5
   update_screen()
   print "Position = %.2f %.2f" % (xpos, ypos)

   return 0

# ------------------------------------------------------------------------------

# get collision detection

def get_collision_detection():

   global collision_detection

   print "Collision Detection =",
   if collision_detection == 0:
      print "Off"
   if collision_detection == 1:
      print "On"

   return 0

# ------------------------------------------------------------------------------

# set collision detection

def set_collision_detection(mode):

   global collision_detection

   collision_detection = mode
   get_collision_detection()

   return 0

# ------------------------------------------------------------------------------

# get obstacle avoidance mode

def get_obstacle_avoidance_mode():

   global obstacle_avoidance_mode

   print "Obstacle Avoidance Mode =",
   print obstacle_avoidance_mode

   return 0

# ------------------------------------------------------------------------------

# set obstacle avoidance mode

def set_obstacle_avoidance_mode(mode):

   global obstacle_avoidance_mode

   obstacle_avoidance_mode = mode
   get_obstacle_avoidance_mode()

   return 0

# ------------------------------------------------------------------------------

# calculate direction and distance

def calculate_direction_and_distance(xpos, ypos):

   print "Calculate Direction And Distance"
   print "Position = %.2f %.2f" % (xpos, ypos)

   deg = get_angle(xpos, ypos)
   dist = get_distance(xpos, ypos)

   print "Direction = %.2f" % deg
   print "Distance = %.2f" % dist

   return 0

# ------------------------------------------------------------------------------

# rotate left

def rotate_left(deg):

   if deg < 0:
      rotate_right(0 - deg)
      return 0

   if deg > 180:
      rotate_right(360 - deg)
      return 0

   print "Rotate Left %.2f degrees" % deg

   amountleft = deg
   while amountleft > 0:
      thistime = min(1, amountleft)
      robot.dir += thistime
      amountleft -= thistime
      robot.dir = normalize(robot.dir)
      update_screen()
      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         if collision_detection == 1:
            print "COLLISION"
            # todo : rotate right ?
            break
      clock.tick(30)

   return 0

# ------------------------------------------------------------------------------

# rotate right

def rotate_right(deg):

   if deg < 0:
      rotate_left(0 - deg)
      return 0

   if deg > 180:
      rotate_left(360 - deg)
      return 0

   print "Rotate Right %.2f degrees" % deg

   amountleft = deg
   while amountleft > 0:
      thistime = min(1, amountleft)
      robot.dir -= thistime
      amountleft -= thistime
      robot.dir = normalize(robot.dir)
      update_screen()
      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         if collision_detection == 1:
            print "COLLISION"
            # todo : rotate left ?
            break
      clock.tick(30)

   return 0

# ------------------------------------------------------------------------------

# rotate to direction

def rotate_to_direction(dir):

   print "Rotate To %.2f degrees" % dir

   dir = normalize(dir)
   deg = dir - robot.dir
   deg = normalize(deg)

   if deg < 0:
      rotate_right(0 - deg)

   if deg > 0:
      rotate_left(deg)

   return 0

# ------------------------------------------------------------------------------

# rotate to position

def rotate_to_position(xpos, ypos):

   print "Rotate To Position %.2f %.2f" % (xpos, ypos)

   deg = get_angle(xpos, ypos)

   if deg < 0:
      rotate_right(0 - deg)

   if deg > 0:
      rotate_left(deg)

   return 0

# ------------------------------------------------------------------------------

# move forward

def move_forward(dist):

   if dist < 0:
      move_backward(0 - dist)
      return 0

   print "Move Forward %.2f units" % dist

   d = robot.dir
   rad = d * pi / 180

   amountleft = dist
   while amountleft > 0:
      thistime = min(2, amountleft)
      robot.x -= thistime * sin(rad)
      robot.y -= thistime * cos(rad)
      amountleft -= thistime
      robot.rect.center = (robot.x, robot.y)
      update_screen()
      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         if collision_detection == 1:
            print "COLLISION"
            # todo : move backward ?
            break
      clock.tick(30)

   return 0

# ------------------------------------------------------------------------------

# move backward

def move_backward(dist):

   if dist < 0:
      move_forward(0 - dist)
      return 0

   print "Move Backward %.2f units" % dist

   d = robot.dir
   rad = d * pi / 180

   amountleft = dist
   while amountleft > 0:
      thistime = min(2, amountleft)
      robot.x += thistime * sin(rad)
      robot.y += thistime * cos(rad)
      amountleft -= thistime
      robot.rect.center = (robot.x, robot.y)
      update_screen()
      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         if collision_detection == 1:
            print "COLLISION"
            # todo : move forward ?
            break
      clock.tick(30)

   return 0

# ------------------------------------------------------------------------------

# move to position

def move_to_position(xpos, ypos):

   print "Move To Position", xpos, ypos

   goal.rect.x = xpos - 5
   goal.rect.y = ypos - 5

   update_screen()

   get_obstacle_avoidance_mode()

   if obstacle_avoidance_mode == 0:
      obstacle_avoidance_mode_0(xpos, ypos)

   if obstacle_avoidance_mode == 1:
      obstacle_avoidance_mode_1(xpos, ypos)

   if obstacle_avoidance_mode == 2:
      obstacle_avoidance_mode_2(xpos, ypos)

   return 0

# ------------------------------------------------------------------------------

# move to goal

def move_to_goal():

   print "Move To Goal"

   xpos = goal.rect.x + 5
   ypos = goal.rect.y + 5

   get_obstacle_avoidance_mode()

   if obstacle_avoidance_mode == 0:
      obstacle_avoidance_mode_0(xpos, ypos)

   if obstacle_avoidance_mode == 1:
      obstacle_avoidance_mode_1(xpos, ypos)

   if obstacle_avoidance_mode == 2:
      obstacle_avoidance_mode_2(xpos, ypos)

   return 0

# ------------------------------------------------------------------------------

# obstacle avoidance mode 0

def obstacle_avoidance_mode_0(xpos, ypos):

   dist = get_distance(xpos, ypos)
   if dist < 5:
      print "ALREADY THERE"
      return 0

   rv = rotate_to_position(xpos, ypos)
   if rv == 1:
      print "ERROR"
      return 1

   move_forward(dist)

   return 0

# ------------------------------------------------------------------------------

# obstacle avoidance mode 1

def obstacle_avoidance_mode_1(xpos, ypos):

   dist = get_distance(xpos, ypos)
   if dist < 5:
      print "ALREADY THERE"
      return 0

   rv = rotate_to_position(xpos, ypos)
   if rv == 1:
      print "ERROR"
      return 1

   print "Move Forward"

   counter = 0

   while True:

      for event in pygame.event.get():
         if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
               return 1

      d = robot.dir
      rad = d * pi / 180
      robot.x -= 2 * sin(rad)
      robot.y -= 2 * cos(rad)
      robot.rect.center = (robot.x, robot.y)

      update_screen()

      clock.tick(30)

      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         print "COLLISION"
         move_backward(5)
         rotate_left(10)
         counter = 0
         continue

      dist = get_distance(xpos, ypos)
      if dist < 5:
         print "WE GOT THERE"
         return 0

      counter += 1
      if counter < 20:
         continue

      counter = 0

      deg = get_angle(xpos, ypos)

      if deg <= -45:
         rotate_right(45)

      if deg > -45 and deg < -0.01:
         rotate_right(0 - deg)

      if deg > 0.01 and deg < 45:
         rotate_left(deg)

      if deg >= 45:
         rotate_left(45)

   return 0

# ------------------------------------------------------------------------------

# obstacle avoidance mode 2

def obstacle_avoidance_mode_2(xpos, ypos):

   dist = get_distance(xpos, ypos)
   if dist < 5:
      print "ALREADY THERE"
      return 0

   rv = rotate_to_position(xpos, ypos)
   if rv == 1:
      print "ERROR"
      return 1

   print "Move Forward"

   counter = 0

   while True:

      for event in pygame.event.get():
         if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
               return 1

      d = robot.dir
      rad = d * pi / 180
      robot.x -= 2 * sin(rad)
      robot.y -= 2 * cos(rad)
      robot.rect.center = (robot.x, robot.y)

      update_screen()

      clock.tick(30)

      if pygame.sprite.spritecollide(robot, obstacles_list, False, pygame.sprite.collide_mask):
         print "COLLISION"
         move_backward(10)
         rotate_left(10)
         counter = 0
         continue

      dist = get_distance(xpos, ypos)
      if dist < 5:
         print "WE GOT THERE"
         return 0

      counter += 1
      if counter < 10:
         continue

      counter = 0

      deg = get_angle(xpos, ypos)

      if deg <= -30:
         rotate_right(30)

      if deg > -30 and deg < -0.01:
         rotate_right(0 - deg)

      if deg > 0.01 and deg < 30:
         rotate_left(deg)

      if deg >= 30:
         rotate_right(30)

   return 0

# ------------------------------------------------------------------------------

# pause program

def pause_program():

   print "Paused ..."
   raw_input("Press Enter to continue ")

   return 0

# ------------------------------------------------------------------------------

# reset

def reset():

   print "Reset"

   robot.x = robot.resetx
   robot.y = robot.resety
   robot.dir = robot.resetd
   robot.rect.center = (robot.x, robot.y)

   goal.rect.x = goal.resetx
   goal.rect.y = goal.resety

   update_screen()

   print "Position = %.2f %.2f" % (robot.x, robot.y)
   print "Direction = %.2f" % robot.dir

   return 0

# ------------------------------------------------------------------------------

# is float

def is_float(str):

   str2 = str.replace(".", "", 1)

   if (str2[0] == "-"):
      tf = str2[1:].isdigit()
   else:
      tf = str2.isdigit()

   return tf

# ------------------------------------------------------------------------------

# get angle

def get_angle(xpos, ypos):

   deltax = xpos - robot.x
   deltay = robot.y - ypos
   angle = atan2(deltay, deltax) / pi * 180
   deg = angle - robot.dir - 90
   deg = normalize(deg)

   if deltax == 0:
      if deltay == 0:
         deg = 0

   return deg

# ------------------------------------------------------------------------------

# get distance

def get_distance(xpos, ypos):

   deltax = xpos - robot.x
   deltay = robot.y - ypos
   dist = sqrt(deltax * deltax + deltay * deltay)

   return dist

# ------------------------------------------------------------------------------

# normalize

def normalize(angle):

   if angle <= -180:
      angle += 360

   if angle > 180:
      angle -= 360

   return angle

# ------------------------------------------------------------------------------

# main program

print "Virtual Robot"

# print "Python version = %s.%s.%s" % sys.version_info[:3]
# print "Pygame version =", pygame.__version__

pygame.display.init()

pygame.display.set_caption("VIRTUAL ROBOT")
screen = pygame.display.set_mode((600, 600))
background = pygame.Surface(screen.get_size())
background.fill((245, 245, 245))
screen.blit(background, (0, 0))

scene = 1

if len(sys.argv) > 1:

   if sys.argv[1] == "/1":
      scene = 1
   if sys.argv[1] == "/2":
      scene = 2
   if sys.argv[1] == "/3":
      scene = 3
   if sys.argv[1] == "/4":
      scene = 4

if scene == 1:

   robot = Robot(100, 100, 270)
   obstacle1 = Obstacle(900, 900, 0, "object1.png")
   obstacle2 = Obstacle(900, 900, 0, "object1.png")
   goal = Goal(900, 900)
   collision_detection = 0
   obstacle_avoidance_mode = 0

if scene == 2:

   robot = Robot(200, 300, 270)
   obstacle1 = Obstacle(300, 300, 0, "object3.bmp")
   obstacle2 = Obstacle(900, 900, 0, "object1.png")
   goal = Goal(400, 300)
   collision_detection = 0
   obstacle_avoidance_mode = 0

if scene == 3:

   robot = Robot(300, 300, 0)
   obstacle1 = Obstacle(120, 450, 150, "object1.png")
   obstacle2 = Obstacle(460, 120, 60, "object1.png")
   goal = Goal(50, 550)
   collision_detection = 1
   obstacle_avoidance_mode = 1

if scene == 4:

   robot = Robot(140, 460, 325)
   obstacle1 = Obstacle(150, 450, 50, "object2.png")
   obstacle2 = Obstacle(450, 150, 250, "object2.png")
   goal = Goal(450, 160)
   collision_detection = 1
   obstacle_avoidance_mode = 1

all_sprites_list = pygame.sprite.Group()
all_sprites_list.add(robot)
all_sprites_list.add(obstacle1)
all_sprites_list.add(obstacle2)
all_sprites_list.add(goal)

obstacles_list = pygame.sprite.Group()
obstacles_list.add(obstacle1)
obstacles_list.add(obstacle2)

update_screen()

clock = pygame.time.Clock()

delay = 20
interval = 20
pygame.key.set_repeat(delay, interval)

print "User Input Mode"
print "Enter ? for Help"

while True:

   input = raw_input("> ")
   words = input.split()

   # ?

   if input == '?':
      help()

   # keyboard

   if input == 'keyboard':
      keyboard_mode()

   # mouse 1

   if input == 'mouse 1':
      mouse_mode_1()

   # mouse 2

   if input == 'mouse 2':
      mouse_mode_2()

   # rs FILE

   for not_used in range(0, 1):
      if len(words) != 2:
         break
      if words[0] != 'rs':
         break
      run_script(words[1])

   # check for commands that can be requested
   # from user input mode or from a script

   do_input(input)

   # q

   if input == 'q':
      break

pygame.quit()

print "Goodbye"

